//CTracker.h++
//Chris Bentzel
//Sep. 97

/***************************************************************************
 * 
 * Copyright 1997  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/


/////////////////////////////////////////////////////////////////
////
////   Pre-processor directives and includes
////
/////////////////////////////////////////////////////////////////
#ifndef _C_TRACKER_HDR_
#define _C_TRACKER_HDR_

//#define C_TRACKER_DEBUG

#include "CVisStateHandler.h++"
#include "CBlob.h++"
#include "GRAYimage.h++"
#include "tdisplay.h++"



/////////////////////////////////////////////////////////////////
////
////   trackInit : data structure for initializing tracker
////
/////////////////////////////////////////////////////////////////
typedef struct trackInit
{
  int imWidth;
  int imHeight;
  int minNumPixels;
  int auxMinNumPixels;
  int minCubePixels;
  float maxError;
  float depthFact;
  int verboseFlag;
  int niftyHacksFlag;
};


typedef struct blobStage
{
  float lowerCentroid[3];
  float upperCentroid[3];
  float width;
};

//Tracker holds a queue of the past couple sequences
const int TRACKER_QUEUE_LENGTH = 30;
const int NUMBER_OF_CUBES = 3;
const int NUMBER_OF_PEOPLE = 1;

//The Blob cache will have nasty data types
typedef CBlob* pCBlob;
typedef pCBlob* blobArray;

typedef CVisState* pCVisState;



/////////////////////////////////////////////////////////////////
////
////   CTracker declaration
////
/////////////////////////////////////////////////////////////////
class CTracker {
  
public:

  //constructor
  CTracker(trackInit& initialData);
  
  ~CTracker();

  //addNewImage:
  //Called when a new image is gathered
  int addNewImage(GRAYimage& Im);

  //sendImageSection: sends area around bounding box
  void sendImageSection();

  void toggleGatherUpperHalfStats();
  void toggleGatherWholeBodyStats();

  void togglePauseImage();
  
protected:

  //performs the connectivity
  void connectivityPass();
  
  //extract valid blobs: places only the good blobs in the right place
  void extractValidBlobs();

  //sorts the blobs in order
  void sortBlobsBySize();
  
  //finds which detectors are on and reports
  void checkDetectors();
  
  //places the blobs with the human
  void assignBlobsToHuman();
  
  //places the blobs with the cubes
  void assignBlobsToCubes();

  //figures out whether the stopping gestures have happened
  void calculateStoppingGestures();
  
  //sets the visState based on the previous two command
  void setVisState();
  
  //prepare display image: wipes it black
  void prepareDisplayImage();
  
  //draws a cross at a given image centroid
  void drawCross(float* centroid, int crossType);
  
  //draw a Bounding Box given two corners
  void drawBBox(int* bbox, int bboxType);
  
  //isSitting returns 1 if the person is sitting,
  void isSitting();

  //converts image coordinates to world coordinates using a wide variety of
  //well-researched hacks.
  void imageToStage(float* blobCentroid, int* boundingBox, blobStage* blobCoord);

  int hysteresis(int newValue, int oldValue, int& currCount, int maxCount);

  //returns malhanobis distance
  float malhanobisDistance(float* data, float* mean, float** invCov, int numFeatures);

  float malhanobisDistance(int whichTest);
  
  void calcHalfBodyStats();

  //timeStampValue used for ordering in array
  int timeStampValue;
  
  //index: the current frame in the blobArrayCache
  int index;
  int lastIndex;

  //blobArrayCace holds onto variable length arrays of
  //blobs over the past n frames
  CBlob** blobArrayCache[TRACKER_QUEUE_LENGTH];

  //currBlobArray-> holds the blobs for the current frame
  CBlob** currBlobArray;
  
  //number Of Blobs tells how many valid blobs
  //were detected for a certain frame
  int numberOfBlobs[TRACKER_QUEUE_LENGTH];
  int numberOfPeople[TRACKER_QUEUE_LENGTH];

  //detector things

  //cubes
  int cubeDetectorNumber;
  int cubeDetectorDetecting[TRACKER_QUEUE_LENGTH];

  //stopping1
  int stopping1DetectorNumber;
  int stopping1DetectorDetecting[TRACKER_QUEUE_LENGTH];
  int stopping1DetectorEvent[TRACKER_QUEUE_LENGTH];

  //stopping2
  int stopping2DetectorNumber;
  int stopping2DetectorDetecting[TRACKER_QUEUE_LENGTH];
  int stopping2DetectorEvent[TRACKER_QUEUE_LENGTH];

  //stopping3
  int stopping3DetectorNumber;
  int stopping3DetectorDetecting[TRACKER_QUEUE_LENGTH];
  int stopping3DetectorEvent[TRACKER_QUEUE_LENGTH];
  
  //up
  int upDetectorNumber;
  int upDetectorDetecting[TRACKER_QUEUE_LENGTH];
  int upDetectorEvent[TRACKER_QUEUE_LENGTH];

  //down
  int downDetectorNumber;
  int downDetectorDetecting[TRACKER_QUEUE_LENGTH];
  int downDetectorEvent[TRACKER_QUEUE_LENGTH];

  //reaching
  int reachingDetectorNumber;
  int reachingDetectorDetecting[TRACKER_QUEUE_LENGTH];
  
  int seatedDetectorNumber;
  int seatedDetectorDetecting[TRACKER_QUEUE_LENGTH];
  int seatedDetectorEvent[TRACKER_QUEUE_LENGTH];
  
  //hysteresis counters
  int maxNumberOfPeopleHysteresisCount;
  int numberOfPeopleHysteresisCount;
  int maxSeatingHysteresisCount;
  int seatingHysteresisCount;
  int seatingHysteresisWacky;
  int crossHysteresisCountUp;
  int maxCrossHysteresisCountUp;
  int crossHysteresisCountDown;
  int maxCrossHysteresisCountDown;
  int hangingHysteresisCountUp;
  int maxHangingHysteresisCountUp;
  int hangingHysteresisCountDown;
  int maxHangingHysteresisCountDown;
  
  //Values for the memory
  CCube* cubeArrayCache[TRACKER_QUEUE_LENGTH][NUMBER_OF_CUBES];
  CPerson* personArrayCache[TRACKER_QUEUE_LENGTH][NUMBER_OF_PEOPLE];

  //blobs for humans and cubes
  int* humanBlobs[TRACKER_QUEUE_LENGTH][NUMBER_OF_PEOPLE];
  int numberOfHumanBlobs[TRACKER_QUEUE_LENGTH][NUMBER_OF_PEOPLE];
  int* cubeBlobs[TRACKER_QUEUE_LENGTH][NUMBER_OF_CUBES];
  int numberOfCubeBlobs[TRACKER_QUEUE_LENGTH][NUMBER_OF_CUBES];

  //quick data to store for each
  float humanImageCentroids[TRACKER_QUEUE_LENGTH][NUMBER_OF_PEOPLE][2];
  float cubeImageCentroids[TRACKER_QUEUE_LENGTH][NUMBER_OF_CUBES][2];
  int humanImageBBoxes[TRACKER_QUEUE_LENGTH][NUMBER_OF_PEOPLE][4];
  int cubeImageBBoxes[TRACKER_QUEUE_LENGTH][NUMBER_OF_CUBES][4];
  int humanImageArea[TRACKER_QUEUE_LENGTH][NUMBER_OF_PEOPLE];
  int cubeImageArea[TRACKER_QUEUE_LENGTH][NUMBER_OF_CUBES];
  eigenStuff humanImageEigenStuff[TRACKER_QUEUE_LENGTH][NUMBER_OF_PEOPLE];
  eigenStuff halfHumanImageEigenStuff[TRACKER_QUEUE_LENGTH][NUMBER_OF_PEOPLE];
  eigenStuff cubeImageEigenStff[TRACKER_QUEUE_LENGTH][NUMBER_OF_CUBES];
  huValues humanImageHuValues[TRACKER_QUEUE_LENGTH][NUMBER_OF_PEOPLE];
  huValues halfHumanImageHuValues[TRACKER_QUEUE_LENGTH][NUMBER_OF_PEOPLE];
  huValues cubeImageHuValues[TRACKER_QUEUE_LENGTH][NUMBER_OF_CUBES];
  
  //A temporary value holding the number of
  //valid blobs for the current frame
  int numberOfValidBlobs;
  
  //A temporary value holding the total
  //number of blobs (valid and invalid) for the
  //current frame
  int numberOfValidAndInvalidBlobs;

  //Image data:
  //mostRecentImage : Current image from background subtraction module.
  //mostRecentImageSection : Small section of the image for displaying
  //                         on television during Scene 2
  //blobMap: labels
  GRAYimage mostRecentImage;
  GRAYimage lastRecentImage;
  GRAYimage mostRecentImageSection;
  GRAYimage blobMap;
  RGBAimage RGBIm;
  
  //bboxSet : used to determine how to transfer image section
  int bboxSet;

  //width and height of the image
  int width;
  int height;

  //minNumberOfPixels: counter for largest value
  //The name is a little confusing: the value this is initially set to
  //is the minimum other blobs are checked against
  int minNumberOfPixels;
  int minCubePixels;
  
  int niftyHacks;
  
  //auxiliaryMinNumberOfPixels : the minimum number
  //of pixels that a disconnected blob should have in order
  //to be slurped up by the largest blob
  int auxiliaryMinNumberOfPixels;
  
  //maxAllowedError: Determines how strong
  //the pull is for the main blob
  //towards disconnected blobs. The larger
  //the error, the stronger the pull.
  float maxAllowedError;

  //depthFactor: used for calculating distance based
  //on lower limit of bounding box
  float depthFactor;

  //test case
  int writeSection;

  CDetector* genericDetector;
  
  //Displays extra debugging information
  int verbose;

  //should we gather statistics?
  int gatherUpperHalfStats;
  int gatherWholeBodyStats;
  
  //The XDisplay for the current image and process
  TDisplay* disp;

  //pause the image
  int pauseImage;
};

#endif //_C_TRACKER_HDR_
  
  
  












